(*
This file is part of the Kind verifier

* Copyright (c) 2007-2010 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

open Types
open Flags
open Exceptions
open Channels
open Globals
open Mpi
open Util

exception EQUIV_CLS_STABLE
exception SOLVER_ERROR

let solver = Globals.my_solver
let toss x = () (* toss outputs *)

(********************************************************************)
let pr_func b_eqs b_imps i_eqs i_imps depth =
  pr "-- "; 
  print_int ((List.length b_imps) + 2*(List.length b_eqs) + 2*(List.length i_eqs) + (List.length i_imps) ) ;
  pr " implications. \n";
  pr "assert";
  let pos_l = Util.n_to_m 1 depth in
    List.iter (fun x-> pr " (true -> (") pos_l ; 
    List.map(fun x -> pr (Imp_graph.eq_org x); pr " \nand ";) b_eqs;
    List.map(fun x -> pr (Imp_graph.imp_org x); pr " \nand ";) b_imps;
    List.map(fun x -> pr (Imp_graph.int_eq_org x); pr " \nand ";) i_eqs;
    List.map(fun x -> pr (Imp_graph.int_imp_org x); pr " \nand ";) i_imps;
    pr " true ";
    List.iter (fun x-> pr "))") pos_l ; 
    pr ";\n" ;;


 

  let pr_out b_eqs b_imps i_eqs i_imps depth fn =
   file_plus_invariant := open_out_gen [Open_append] 0644 (fn^".inv");
  output_string !file_plus_invariant "assert";
  let pos_l = Util.n_to_m 1 depth in
    List.iter (fun x-> output_string !file_plus_invariant " (true -> (") pos_l ;
    List.map(fun x -> output_string !file_plus_invariant (Imp_graph.eq_org x); output_string !file_plus_invariant " \nand ";) b_eqs;
    List.map(fun x -> output_string !file_plus_invariant (Imp_graph.imp_org x); output_string !file_plus_invariant " \nand ";) b_imps;
    List.map(fun x ->  output_string !file_plus_invariant (Imp_graph.int_eq_org x);  output_string !file_plus_invariant " \nand ";) i_eqs;
    List.map(fun x ->  output_string !file_plus_invariant (Imp_graph.int_imp_org x);  output_string !file_plus_invariant " \nand ";) i_imps;
     output_string !file_plus_invariant " true ";
    List.iter (fun x->  output_string !file_plus_invariant "))") pos_l ; 
     output_string !file_plus_invariant ";\n" ;
  output_string !file_plus_invariant "tel." ;;   



let print_invariant b_eqs b_imps i_eqs i_imps depth=
  let buf = Buffer.create 100 in
     Buffer.add_string buf "(define ASSERTIONS :: (-> _nat bool) (lambda ( _M::nat) (ite (= _M _base) true " ;
    Buffer.add_string buf "( and";
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.equ_invariant x);) b_eqs;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.imp_invariant x);) b_imps;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.int_eq_invariant x);) i_eqs;
    List.map(fun x ->   Buffer.add_string buf (Imp_graph.int_imp_invariant x);) i_imps; 
     Buffer.add_string buf "))))";
     (Buffer.contents buf);;   

let main fn = 
  let defdoc,maxdepth,def_hash,no_stateful,pvars = Defgen.start fn in

let cc = solver#get#cc in
print_to_user(cc^": MAIN. \n"); 

  (* generate defs for new vars of equalities. Yeting  *)
  let all_terms = Lus_convert_yc.get_all_def_terms () in
  let bool_sub_exprs, int_sub_exprs, float_sub_exprs = 
    Sub_exprs.sub_exprs all_terms Expr_util.filter_sub_exprs in

  (* generate new vars defs. Yeting *)
  let new_bool_var_doc = New_vars.bool_newvar_defs bool_sub_exprs in
  let new_int_var_doc =  New_vars.int_newvar_defs int_sub_exprs in

  (* Teme: Creating the equiv classes *)
  (* Teme: Boolean invariants *)
  let bool_imp_graph =     
    new Imp_graph.bool_imp_graph (
      if (!Flags.invariant_bool) 
      then (New_vars.get_bool_nvrs ())
      else []
    )
  in

  (*Teme: Integer invariants *)
  let ints_partial_order = 
    new Partial_order.partial_order (
      if (!Flags.invariant_int) 
      then New_vars.get_int_nvrs ()
      else []
    ) 
  in
  (* end of new vars *)
    
  Kind_support.print_stat_headers();

  let cc = solver#get#cc in
  let nvar = solver#get#k_position_string in
  let add = Kind_support.get_add() in
  let nstart = solver#get#step_base_string in
  let startdepth = maxdepth + 1 in
  let from_solver_ch, from_checker_ch, from_solver2_ch = Kind_support.setup_channels() in

  (* print definitions *)
  Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (defdoc^"\n");

 (* print definitions of new vars .  Yeting *)
    print_to_both solver#get#push_string; (***** PUSH all new var defs*****)
    Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (new_bool_var_doc^"\n");
    Kind_support.print_to_both_limited from_solver_ch from_solver2_ch from_checker_ch (new_int_var_doc^"\n");
 (* end of print definitions of new vars *)

  if !checker_mode then print_to_checker solver#get#checker_setup_string;

  print_to_user (cc^"INV GEN PROC: Begin induction:\n");
  (* print initial case *)


  print_to_both solver#get#push_string; (***** PUSH *****)
  Kind_support.print_initialization def_hash startdepth add from_checker_ch;

  (* print new variables assertions. Yeting  *)
  print_to_both solver#get#push_string; (***** PUSH *****)
  print_to_solver (New_vars.mk_nvr_eq_cmds (NUM 0));
  (* print new variables assertions. Yeting  *)
 

  let k_inc = ref startdepth in
  let maxsteps = max !maxloops (startdepth+1) in
  let firststep = ref true in

  Kind_support.set_last_level_asserted (startdepth-1);


  let rec filter_imp_equiv_classes_one_round stepn_positions =
    print_to_solver (cc ^ "INV GEN PROC:Filter \n");
    print_to_solver solver#get#push_string; 

    if (!Flags.invariant_bool)
    then 
      List.iter 
      (fun x-> print_to_solver x; )
      (bool_imp_graph#doc_of_stepn_asserts stepn_positions);

    if (!Flags.invariant_int)
    then
    List.iter 
      (fun x-> print_to_solver x; )
      (ints_partial_order#doc_of_stepn_asserts stepn_positions);
     
    let bool_imp_doc = 
      if !Flags.invariant_bool
      then (bool_imp_graph#doc_of_asserts ())
      else [] 
    in
 
    let int_imp_doc = 
      if !Flags.invariant_int
      then (ints_partial_order#doc_of_asserts ())
      else []
    in
   
    let imp_equiv_doc = Expr_util.mk_not_ands (bool_imp_doc@int_imp_doc)
    in
      if ("" = imp_equiv_doc) 
      then false
      else (
      print_to_solver imp_equiv_doc;
      print_to_solver (
        Lus_convert_yc.simple_command_string (
          QUERY(F_TRUE)));
      
      
      print_to_solver solver#get#done_string;


    let out1 = solver#get#get_output from_solver_ch in
      if (solver#get#result_is_unsat out1) 
      then(
        print_to_solver solver#get#pop_string; (* PUSH 2 *)
	true)
    else if (solver#get#result_is_sat out1) 
    then(
      let simulation_value_hash = 
	solver#get#get_simulation_value_hash out1 
	  print_to_solver from_solver_ch  in

      let changed = 
	if (!Flags.invariant_bool && !Flags.invariant_int)
	then (
	  let ra = bool_imp_graph#set_new_graph simulation_value_hash in
	  let rb = ints_partial_order#sort simulation_value_hash in
	    if (ra || rb) then true else false) 
	else if (!Flags.invariant_bool)
	then (bool_imp_graph#set_new_graph simulation_value_hash)
	else if (!Flags.invariant_int)
	then (ints_partial_order#sort simulation_value_hash )
	else failwith "Error"
      in
	print_to_solver solver#get#pop_string;
	if (changed ) 
	then ( 
	  (*pr "changed\n"; *)
	  filter_imp_equiv_classes_one_round stepn_positions 
	)
        else  false 
    )
    else ( (* error *)
      if (solver#get#result_is_error out1) then
        print_to_user_final ((Str.matched_string out1)^"\n");
      print_to_user (cc^"INV GEN PROC: SOLVER OUTPUT: "^out1);
      Globals.error := true;
      raise SOLVER_ERROR
    )
      )
  in
    (* some vars to control loops Yeting *)
  let cur_depth = ref maxdepth  in
  let last_changed = ref !cur_depth in
    (* end of vars *)

  let final_position = 0 in
  (* this does not include 0 : 1 to k-1 *)

let stable = ref false in


while not !stable do
  (*let flag =  probe_nonblocking master_proc any_tag comm_world in
  if flag = 0 then
    begin*)
     
  while  not !Globals.stop do
     print_to_user (cc^"INV GEN PROC: - BASE LOOP  \n");
    (* Checking BASE *)
    print_to_solver solver#get#push_string; (***** PUSH *****)

    (* print the assertion of equiv classes.  Yeting *) 
    let bool_imp_doc =
      if (!Flags.invariant_bool)
      then  bool_imp_graph#doc_of_asserts()
      else []
    in

    let int_imp_doc =
      if (!Flags.invariant_int)
      then ints_partial_order#doc_of_asserts()
      else []
    in


    let equiv_imp_doc = Expr_util.mk_not_ands (bool_imp_doc@int_imp_doc) in
      print_to_solver equiv_imp_doc;
   
   if (equiv_imp_doc = "")
   then (pr_func [] [] [] [] 1; exit 0 );

    let base_num = !cur_depth * add in
     print_to_user (cc^"INV GEN PROC: Asserting base case\n");
    let assert_base = (F_EQ(POSITION_VAR(nstart),NUM(base_num),L_INT)) in
      print_to_solver 
	(Lus_convert_yc.simple_command_string (ASSERT assert_base));
      
      print_to_solver (
        Lus_convert_yc.simple_command_string (
          QUERY(F_TRUE)));

    (* BASE check *)
     print_to_user (cc^"INV GEN PROC: Checking base case\n");
    print_to_solver solver#get#done_string;
    (* finished BASE check *)
    
    let out1 = solver#get#get_output from_solver_ch in

    if (solver#get#result_is_unsat out1) then
      begin (* base valid *)
	print_to_user (cc^"INV GEN PROC: Valid base case - K NOT Stable\n");
        print_to_solver solver#get#pop_string; (* PUSH 2 *)

        if (!cur_depth ) > (!last_changed )  then 
	  print_to_user (cc^"INV GEN PROC: Valid base case - K Stable\n");
	  stable := true;
	Globals.stop := true;
	
        Kind_support.persistent_step_asserts_concrete 
          def_hash startdepth add (!cur_depth+1) from_checker_ch;
	cur_depth := !cur_depth + 1 ; 
	
      end (* base valid *)
    else if (solver#get#result_is_sat out1) then
      begin (* base invlaid *)
        print_to_user (cc^"INV GEN PROC: Invalid base case - K NOT Stable\n");
        let basestep = 
          if !Globals.firsttime then
            BASE_INIT
          else
            BASE
        in
    let foundvars = solver#get#get_countermodel out1 print_to_solver from_solver_ch in
	  last_changed := !cur_depth;
	  (* get the values of nvrs. Yeting *) 
    let simulation_value_hash = solver#get#get_simulation_value_hash out1 print_to_solver from_solver_ch  in

    let changed  = 
	  if (!Flags.invariant_int && !Flags.invariant_bool)
	  then (
	    let ra = bool_imp_graph#set_new_graph simulation_value_hash in
	    let rb = ints_partial_order#sort simulation_value_hash in
	      if (ra || rb ) then true else false )
	  else if (!Flags.invariant_bool)
	  then (bool_imp_graph#set_new_graph simulation_value_hash)
	  else if (!Flags.invariant_int)
	  then (ints_partial_order#sort simulation_value_hash)
	  else failwith "Error"
	in
	  
        print_to_solver solver#get#pop_string;
        print_to_solver solver#get#safe_pop; (* PUSH 2 *)
          begin
	    Globals.stop := not changed ;
	    if(!Globals.stop ) 
	    then (
	      pr (Expr_util.mk_not_ands (bool_imp_graph#doc_of_asserts() ));
	      pr "No changes while sat.\n";
	      failwith "global stop"; )
	  (*Globals.stop := true*)
          end
      end (* base invalid *)
    else
      begin
        (* BASE check = ERROR *)
        if (solver#get#result_is_error out1) then
          print_to_user_final ((Str.matched_string out1)^"\n");
        print_to_user (cc^"SOLVER OUTPUT: "^out1);
        Globals.error := true;
        Globals.stop := true
      end;
  done; (* main loop *)

	  print_to_user(cc^"INV GEN PROC: STABLE K\n");
	  stable := true;
      
done; (* main stable*)
 
    
      	  print_to_user(cc^"INV GEN PROC: STABLE K - DOING INDUCTION.\n");

      print_to_solver solver#get#pop_string; 
      print_to_solver solver#get#pop_string; 
      (* Print new var defs *)
      print_to_both solver#get#push_string; (***** PUSH *****)
      print_to_solver (New_vars.mk_nvr_eq_cmds (POSITION_VAR(nvar)));
      (* assert new base *)
      let base_num = !cur_depth * add in
      let assert_base = (F_EQ(POSITION_VAR(nstart),NUM(base_num),L_INT)) in
	print_to_solver 
	  (Lus_convert_yc.simple_command_string (ASSERT assert_base));
      (* print defs *)

      let stepn_positions =  
        List.map 
	  (fun x -> MINUS(POSITION_VAR(nvar),NUM(x))) 
	  (Util.n_to_m 1 (!cur_depth - 1))  in 

      List.iter (fun x -> 
	Kind_support.def_assert_both1 def_hash "DEF" 
          [PLUS(POSITION_VAR(nvar),NUM(x*add))]
          !cur_depth from_checker_ch;) (Util.n_to_m 0 (!cur_depth - 1) ) ;

      if filter_imp_equiv_classes_one_round stepn_positions
      then ( 
	    let b_eqs,b_imps =
	      if (!Flags.invariant_bool)
	      then bool_imp_graph#eq_imp_graph()
	      else [],[]
	    in
	    let int_eqs,int_imps =
	      if (!Flags.invariant_int)
	      then ints_partial_order#eq_imp()
	      else [],[]
	    in
	    Globals.inv_gen_time_stop := (wtime());	
	    
	      print_to_user("INV GEN PROC: Sending Invariant to STEP PROC. \n");		
	
	      let invariant = print_invariant b_eqs b_imps int_eqs int_imps maxdepth in  
              let s_invariant = INV_FULL(invariant) in 
		send s_invariant step_proc 89 comm_world;
		   
      )
    else (
      pr "assert true;\n";
    );

